#include <xil_exception.h>
#include <xstatus.h>
#include <pm_api_sys.h>
#include <libeswpm/test_aux.h>

extern void __attribute__((weak)) *_vector_table;
static XScuGic GicInst;
static XIpiPsu IpiInst;

#ifdef __aarch64__
#define PU_NODE		NODE_APU_0
#define SRAM_0		NODE_OCM_BANK_0
#define SRAM_1		NODE_OCM_BANK_1
#define SRAM_2		NODE_OCM_BANK_2
#define SRAM_3		NODE_OCM_BANK_3
#else
#define PU_NODE		NODE_RPU_0
#define SRAM_0		NODE_TCM_0_A
#define SRAM_1		NODE_TCM_0_B
#define SRAM_2		NODE_TCM_1_A
#define SRAM_3		NODE_TCM_1_B
#endif

/**
 * prepare_suspend() - save sontext and request suspend
 */
static void prepare_suspend(void)
{
	XStatus status;

	status = XPm_SelfSuspend(PU_NODE, MAX_LATENCY, 0, (u64)&_vector_table);
	HALT_IF_ERR(status, "in self suspend call for %s\r\n", SHOW_DEFINE(PU_NODE));

#ifdef __aarch64__
	status = XPm_SetRequirement(NODE_L2, 0, 0, REQUEST_ACK_NO);
	HALT_IF_ERR(status, "in set requirement for NODE_L2\r\n");
#endif
	status = XPm_SetRequirement(SRAM_0, PM_CAP_CONTEXT, 0, REQUEST_ACK_NO);
	HALT_IF_ERR(status, "in set requirement for %s\r\n", SHOW_DEFINE(SRAM_0));

	status = XPm_SetRequirement(SRAM_1, PM_CAP_CONTEXT, 0, REQUEST_ACK_NO);
	HALT_IF_ERR(status, "in set requirement for %s\r\n", SHOW_DEFINE(SRAM_1));

	status = XPm_SetRequirement(SRAM_2, PM_CAP_CONTEXT, 0, REQUEST_ACK_NO);
	HALT_IF_ERR(status, "in set requirement for %s\r\n", SHOW_DEFINE(SRAM_2));

	status = XPm_SetRequirement(SRAM_3, PM_CAP_CONTEXT, 0, REQUEST_ACK_NO);
	HALT_IF_ERR(status, "in set requirement for %s\r\n", SHOW_DEFINE(SRAM_3));

#ifdef __aarch64__
	status = XPm_SetRequirement(NODE_DDR, PM_CAP_CONTEXT, 0, REQUEST_ACK_NO);
	HALT_IF_ERR(status, "in set requirement for NODE_DDR\r\n");

	GicSuspend(&GicInst);
#endif
}

int main(void)
{
	enum XPmBootStatus BootStatus;

	/* Get boot status */
	BootStatus = XPm_GetBootStatus();
	if (PM_INITIAL_BOOT == BootStatus) {
		pm_print("INITIAL BOOT\r\n");

		/* Initialize GIC, IPIs, and Xilpm */
		PmTestInit(&GicInst, &IpiInst);

		/* Let other masters in the system know I'm ready */
		SyncDeclareReady();
	} else if (PM_RESUME == BootStatus) {
		SyncDeclareReady();
		pm_print("RESUMED\r\n");
#ifdef __aarch64__
		int Status = XST_FAILURE;

		/* Restore GIC state */
		Status = GicResume(&GicInst);
		HALT_IF_ERR(Status, "failed to resume GIC\r\n");
#endif
		/* Enable interrupts in cpsr */
		Xil_ExceptionEnable();
	} else {
		pm_print("ERROR cannot identify boot reason\r\n");
	}

	/* Wait to receive init suspend request from PMU */
	IpiWaitForInitSuspend();
	if (pm_susp.reason != SUSPEND_REASON_PU_REQ) {
		pm_print("Unexpected suspend reason %u\r\n", pm_susp.reason);
		goto fail;
	}

	/* Initiate suspend */
	prepare_suspend();
	XPm_SuspendFinalize();

fail:
	/* Should never execute code below */
	XIL_FAIL(1);
	return XST_FAILURE;
}
